/*  Written by Julian Grizzard
 *  September 2, 2002
 *  Contact: gtg993j@mail.gatech.edu
 *
 *  Please contact me if you decide to publish any of this code.  Thanks!
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <errno.h>
#include "client.h"

/* Local prototypes */
int nextLine(); /* point finger to next server line recving it if neccessary */
int nextTok(); /* Point finger to next token in recv message */

/* Global vars */
int server;        /* Socket for server */
char *buf;         /* used to recv messages */
int bufferSize = MAX_RECV_MSG; /* current buffer size */
char *finger;      /* used to point to current line */


int initializeWorld(s_world *world, list *robotL, list *homeL, 
                    list *pkgL, char *hostName, int port)
{
   struct sockaddr_in localAddr;
   struct sockaddr_in remoteAddr;
   struct hostent *h; /* Stores info about the server */
   int i;
   

   /* Initialize the world, bots, home list, package list */
   h = NULL;
   i = 0;
   buf = NULL;

   buf = (char *) malloc(bufferSize);
   memset(buf, 0, bufferSize);
   finger = &buf[0];
   if(buf == NULL)
   {
      errno = ENOMEM;
      return -1;
   }

   h = gethostbyname(hostName);
   if(h == NULL)
   {
      errno = EBAD_HOSTNAME;
      return -1;
   }

   remoteAddr.sin_family = h->h_addrtype;
   memcpy((char *) &remoteAddr.sin_addr.s_addr, h->h_addr_list[0], h->h_length);
   remoteAddr.sin_port = htons(port);

   /* Create the socket */
   server = socket(AF_INET, SOCK_STREAM, 0);
   if(server < 0)
   {
      errno = ESOCKET;
      return -1;
   }

   /* Bind any port */
   localAddr.sin_family = AF_INET;
   localAddr.sin_addr.s_addr = htonl(INADDR_ANY);
   localAddr.sin_port = htons(0);
   if(bind(server, (struct sockaddr *) &localAddr, sizeof(localAddr)) < 0)
   {
      errno = EBIND;
      return -1;
   }

   /* Connect to remote server */
   fprintf(stdout, "   Connecting to remote server...");
   if(connect(server, (struct sockaddr *) &remoteAddr, sizeof(remoteAddr)) < 0)
   {
      printf("failed\n");
      errno = ECONNECT;
      return -1;
   }
   printf("established\n");

   /* 1) Send init string and receive server response */
   fprintf(stdout, "   Sending init string %s to server...\n", C_INIT);
   if(sendCommand(C_INIT) < 0)
   {
      errno = ESEND;
      return -1;
   }
   if(nextLine() < 0)
   {
      return -1;
   }

   /* 2) Get the map */
   /* First get the Width and Height: "W H" */
   printf("   Downloading map\n");
   if(sscanf(finger, "%d %d", &world->width, &world->height) < 0)
   {
      errno = EBAD_MAP;
      return -1;
   }
   fprintf(stdout, "   width: %d height: %d\n", world->width, world->height);

   /* allocate space for map and put a wall around the perimeter
    * of the map
    */
   world->map = (char **) malloc((world->width + 2) * sizeof(char *));
   for(i = 0; i < world->width + 2; i++)
   {
      world->map[i] = (char *) malloc((world->height + 2) * sizeof(char));
   }
   for(i = world->height + 1 ; i >= 0; i--)
   {
      int j = 0;
      for(j = 0; j < world->width + 2; j++)
      {
         world->map[j][i] = ' ';
      }
   }
   for(i = 0; i < world->width + 2; i++)
   {
      world->map[i][0] = B_WALL;
      world->map[i][world->height + 1] = B_WALL;
   }
   for(i = 1; i < world->height + 1; i++)
   {
      world->map[0][i] = B_WALL;
      world->map[world->width + 1][i] = B_WALL;
   }

   /* now get all of the rows */
   for(i = 1; i <= world->height; i++)
   {
      int j;
      if(nextLine() < 0)
      {
         return -1;
      }
      for(j = 1; j <= world->width; j++)
      {
         world->map[j][i] = *finger;
         finger++;
         if(world->map[j][i] == B_HOME)
         {
            /* found a home base add to list */
            s_home *theHome;
            theHome = newHome();
            theHome->x = j;
            theHome->y = i;
            addElement(homeL, theHome);
         }
      }
   }


   /* 3) Get player configuration */
   fprintf(stdout, "   Retrieving Player config...\n");
   {
      element *head;
      s_robot *theBot; 
      theBot = newBot();
      //fprintf(stdout, "      adding bot...\n");
      if(addElement(robotL, theBot) < 0)
         fprintf(stderr, "Error adding bot to list\n");
      if(headElement(robotL, &head) < 0)
         fprintf(stderr, "Error getting head item\n");
      if(nextLine() < 0)
      {
         return -1;
      }
      if(sscanf(finger, "%d %d %d", &((s_robot *)head->value)->id, 
                                 &((s_robot *)head->value)->cpty, 
                                 &((s_robot *)head->value)->money) < 0)
      {
         errno = EBAD_CONF;
         return -1;
      }
      fprintf(stdout, "   id: %d cpty %d money %d\n", ((s_robot *)head->value)->id, 
                                 ((s_robot *)head->value)->cpty,  
                                 ((s_robot *)head->value)->money);
   }

   /* 4) Get Initial server response */
   fprintf(stdout, "   Update world for first time\n");
   if(updateWorld(world, robotL, homeL, pkgL) < 0)
   {
      return -1;
   }

   return 0;
}

int nextLine()
{
   int recvSize = 0;
   int offset = 0;
   while(*finger != 0)
   {
      if(*finger == 10)
      {
         finger++;
         if(*finger != 0)
         {
            return 0; /* finger now pointing to next line */
         }
      }
      else
      {
         finger++;
      }
   }
   /* Got to a 0 before finding a new line, so we need 
    * to recveive more from the server.  First, just clear
    * the buffer
    */
   memset(buf, 0, bufferSize);
   offset = 0;
   recvSize = 0;
   do
   {
      if(recvSize == bufferSize)
      {
         /* Increase buffersize */
         free(buf);
         bufferSize = bufferSize * 2;
         buf = (char *) malloc(bufferSize);
         memset(buf, 0, bufferSize);
      }
      recvSize = recv(server, buf, bufferSize, MSG_PEEK);
      if(recvSize < 0)
      {
         errno = ERECV;
         return -1;
      }
   } while(recvSize == bufferSize);
   recvSize = recv(server, buf, bufferSize, 0);
   if(recvSize < 0)
   {
      errno = ERECV;
      return -1;
   }
   finger = &buf[0];
   return 0;
}

int getPackages(list *pkgL, list *robotL)
{
   s_pkg *thePkg;
   int pkgID = 0;
   int pkgX = 0;
   int pkgY = 0;
   int pkgW = 0;
   if(nextLine() < 0)
   {
      return -1;
   }
   //printf("   getting packages...\n");
   printf("\n------SERVER REPSPONSE PACKAGES-----------\n");
   printf("%s", finger);
   printf("----------END SERVER REPSPONSE---------\n\n");
   
   while(sscanf(finger, "%d %d %d %d", &pkgID, &pkgX, &pkgY, &pkgW) == 4)
   {

      thePkg = findPkg(pkgL, pkgID);
      if(thePkg == NULL)
      {
         thePkg = newPkg();
         thePkg->id = pkgID;
         addElement(pkgL, thePkg);
      }
   
      thePkg->x = pkgX;
      thePkg->y = pkgY;
      thePkg->xLoc = botX(robotL);
      thePkg->yLoc = botY(robotL);
      thePkg->weight = pkgW;

      if(nextTok() < 0)
      {
         return 0;
      }
      nextTok();
      nextTok();
      nextTok();

   }

   return 0;
}

int sendCommand(char *command)
{
   char *bufSend;
   int length = 0;
   length = strlen(command);
   bufSend = (char *) malloc(length + 1);
   strncpy(bufSend, command, length);
   bufSend[length] = 10;
   //bufSend[length + 1] = 0;
   printf("++++++++SENDING TO SERVER++++++++++++\n");
   printf("%s", bufSend);
   printf("++++++END SENDING TO SERVER++++++++++++\n");
   if(send(server, bufSend, length + 1, 0) < 0)
   {
      errno = ESEND;
      return -1;
   }
   return 0;
}

int updateWorld(s_world *world, list *robotL, list *homeL, list *pkgL)
{
   s_robot *theBot = NULL;
   int notDone = 1;
   element *fingE;

   /* First Assume all robots are dead and change
    * them accordingly.
    * If they are alive, then they will appear
    * in the server response and we will change
    * their state back to alive.
    */
   if(headElement(robotL, &fingE) == 0)
   {
      do
      {
         ((s_robot *)fingE->value)->state = R_DEAD;
      } while(nextElement(robotL, fingE, &fingE) == 0);
   }

   /* Now Get the server Response */
   if(nextLine() < 0)
   {
      return -1;
   } 
   printf("\n------------SERVER REPSPONSE-----------\n");
   printf("%s", finger);
   printf("----------END SERVER REPSPONSE---------\n\n");
   theBot = NULL;
   /* Parse the response */
   //fprintf(stdout, "Parse the server response: %s\n", finger);
   while(notDone)
   {
      int robotID = 0;
      int xCord = 0;
      int yCord = 0;
      int thePkg = 0;
      if(*finger == 10)
      {
         notDone = 0;
         break; 
      }
      switch(*finger)
      {
         case A_ID:
            //printf("found bot\n");
            if(sscanf(finger, "#%d", &robotID) <= 0)
            {
               errno = EBAD_WORLD;
               return -1;
            }
            theBot = findBot(robotL, robotID);
            if(theBot == NULL)
            {
               /* Bot doesn't exist so add it to list */
               theBot = newBot();
               addElement(robotL, theBot);
            }
            theBot->id = robotID;
            theBot->state = R_ALIVE;
            break;
         case A_NORTH:
            theBot->y++;
            break;
         case A_SOUTH:
            theBot->y--;
            break;
         case A_EAST:
            theBot->x++;
            break;
         case A_WEST:
            theBot->x--;
            break;
         case A_PICK:
            nextTok();
            if(sscanf(finger, "%d", &thePkg) <= 0)
            {
               errno = EBAD_WORLD;
               return -1;
            }
            addPkg(theBot, pkgL, thePkg);
            break;
         case A_DROP:
            nextTok();
            if(sscanf(finger, "%d", &thePkg) <= 0)
            {
               errno = EBAD_WORLD;
               return -1;
            }
            dropPkg(theBot, pkgL, thePkg);
            break;
         case A_XCORD:
            //printf("got x coord\n");
            nextTok();
            if(sscanf(finger, "%d", &xCord) <= 0)
            {
               errno = EBAD_WORLD;
               return -1;
            }
            theBot->x = xCord;
            break;
         case A_YCORD:
            //printf("got y coord\n");
            nextTok();
            if(sscanf(finger, "%d", &yCord) <= 0)
            {
               errno = EBAD_WORLD;
               return -1;
            }
            theBot->y = yCord;
            break;
         default:
            errno = EBAD_WORLD;
            return -1;
            break;
      }
      //printf("before moving to next token: %s\n", finger);
      if(nextTok() < 0)
      {
         notDone = 0;
         break;
      }
      //printf("after moving to next token: %s\n", finger);
   }
   
   return 0;
}

int nextTok()
{
   while(*finger != 10 && *finger != ' ')
   {
      finger++;
   }
   if(*finger == 10)
   {
      return -1; /* Next token not found */
   }
   while(*finger == ' ')
   {
      finger++;
   }
   if(*finger == 10)
   {
      return -1; /* Next token not found */
   }
   return 0; /* Next token found */
}

