/*  Written by Julian Grizzard
 *  September 2, 2002
 *  Contact: gtg993j@mail.gatech.edu
 *
 *  Please contact me if you decide to publish any of this code.  Thanks!
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include "robot.h"
#include "world.h"
#include "client.h"
#include "ai.h"
#include "linkedList.h"

/* local prototypes */

int printUsage();

int main(int argc, char **argv)
{
   s_world *world;
   list *robotL;
   list *homeL;
   list *pkgL;
   char hostName[MAX_HOST_NAME];
   int port;

   world = (s_world *) malloc(sizeof(s_world));
   world->width = 0;
   world->height = 0;
   world->map = NULL;
   robotL = newList();
   homeL = newList();
   pkgL = newList();
   hostName[0] = 0;
   port = 0;


   /* parse args */
   if(argc < 2)
   {
      printUsage(argv);
      exit(-1);
   }
   strncpy(hostName, argv[1], MAX_HOST_NAME);
   if(sscanf(argv[2], "%d", &port) <= 0)
   {
      printUsage(argv);
      exit(-1);
   }
   
   fprintf(stdout, "Initializing World...\n");
   if(initializeWorld(world, robotL, homeL, pkgL, hostName, port) < 0)
   {
      fprintf(stderr, "Error Initializing World\n");
      exit(-1);
   }

//   printState(world, robotL, homeL, pkgL);

   while(1)
   {
      /* First, Get list of packages at this location */
      fprintf(stdout, "1) Getting list of packages...\n");
      if(getPackages(pkgL, robotL) < 0)
      {
         fprintf(stderr, "Error Getting Packages\n");
         exit(-1);
      }

      /* Second, issue command to server */
      //printState(world, robotL, homeL, pkgL);
      fprintf(stdout, "2) Issuing robot command...\n");
      if(issueCommand(world, robotL, homeL, pkgL) < 0)
      {
         fprintf(stderr, "Error Issuing Command\n");
         exit(-1);
      }

      /* Third, update the world, what happened? */
      fprintf(stdout, "3) Retrieving server response...");
      if(updateWorld(world, robotL, homeL, pkgL) < 0)
      {
         fprintf(stderr, "Error Updating World\n");
         exit(-1);
      }
      //printState(world, robotL, homeL, pkgL);
   } 

   /* shouldn't ever get here, so must be some wierd error */ 
   return -1;
}

int printUsage(char **argv)
{
   fprintf(stderr, "Usage: %s <host> <port>\n", argv[0]);
   return 0;
}

/* Send command to move in direction d
 * where d is N, S, E, or W
 */
int move(char *d, int priority, list *robotL)
{
   element *head;
   char *command;
   char prior[32];
   memset(prior, 0, 32);

   sprintf(prior, "%d ", priority);

   command = NULL;
   command = (char *) malloc(strlen(prior) + sizeof(C_MOVE) + strlen(d) + 3);
   if(command == NULL)
   {
      errno = ENOMEM;
      return -1;
   }

   strncpy(command, prior, strlen(prior) + 1);
   strncat(command, C_MOVE, sizeof(C_MOVE) + 1);
   strncat(command, d, strlen(d) + 1);

   if(sendCommand(command) < 0)
   {
      return -1;
   }

   headElement(robotL, &head);
   ((s_robot *)head->value)->money = ((s_robot *)head->value)->money -  
                                     abs(priority);

   return 0;
}

/* Send command to pick up list
 */
int pick(char *pickList, int priority, list *robotL)
{
   element *head;
   char *command;
   char prior[32];
   memset(prior, 0, 32);

   sprintf(prior, "%d ", priority);

   command = NULL;
   command = (char *) malloc(strlen(prior) + sizeof(C_PICK) + strlen(pickList) + 3);
   if(command == NULL)
   {
      errno = ENOMEM;
      return -1;
   }

   strncpy(command, prior, strlen(prior) + 1);
   strncat(command, C_PICK, sizeof(C_PICK) + 1);
   strncat(command, pickList, strlen(pickList) + 1);

   if(sendCommand(command) < 0)
   {
      free(command);
      return -1;
   }

   headElement(robotL, &head);
   ((s_robot *)head->value)->money = ((s_robot *)head->value)->money -  
                                     abs(priority);

   //printf("freeing\n");
   free(command);
   //printf("freeing done\n");
   return 0;
}

/* Send command to drop list
 */
int drop(char *dropList, int priority, list *robotL)
{
   char *command;
   element *head;
   char prior[32];
   memset(prior, 0, 32);

   sprintf(prior, "%d ", priority);

   command = NULL;
   command = (char *) malloc(strlen(prior) + sizeof(C_DROP) + strlen(dropList) + 3);
   if(command == NULL)
   {
      errno = ENOMEM;
      return -1;
   }

   strncpy(command, prior, strlen(prior) + 1);
   strncat(command, C_DROP, sizeof(C_DROP) + 1);
   strncat(command, dropList, strlen(dropList) + 1);

   if(sendCommand(command) < 0)
   {
      free(command);
      return -1;
   }


   headElement(robotL, &head);
   ((s_robot *)head->value)->money = ((s_robot *)head->value)->money - 
                                     abs(priority);
   free(command);

   return 0;
}

s_robot *newBot()
{
   s_robot *newRobot = NULL;
   newRobot = (s_robot *) malloc(sizeof(s_robot));
   newRobot->id = 0;
   newRobot->x = 0;
   newRobot->y = 0;
   newRobot->money = 0;
   newRobot->cpty = 0;
   newRobot->pkgs = newList();
   newRobot->state = R_DEAD;
   return newRobot;
}

s_robot *findBot(list *robotL, int robotID)
{
   element *botElem;
   if(headElement(robotL, &botElem) < 0)
   {
      return NULL;
   }
   do
   {
      if(((s_robot *)botElem->value)->id == robotID)
      {
         return (s_robot *)botElem->value;
      }
   } while(nextElement(robotL, botElem, &botElem) == 0);
   return NULL;
}

int addPkg(s_robot *robot, list *pkgL, int pkgID)
{
   s_pkg *thePkg;
   
   thePkg = findPkg(pkgL, pkgID);
   thePkg->state = P_CARRIED;
   addElement(robot->pkgs, thePkg);

   robot->cpty = robot->cpty - thePkg->weight;
 
   return 0;
}

int dropPkg(s_robot *robot, list *pkgL, int pkgID)
{
   s_pkg *thePkg;
 
   thePkg = findPkg(robot->pkgs, pkgID);
   if(thePkg == NULL)
   {
      /* We didn't know about this package, 
       * so create it and add to list
       */
      thePkg = newPkg();
      thePkg->id = pkgID;
      addElement(pkgL, thePkg);
   }
   removeElement(robot->pkgs, thePkg);
   thePkg->state = P_GROUNDED;
   thePkg->xLoc = robot->x;
   thePkg->yLoc = robot->y;
   if(thePkg->xLoc == thePkg->x && thePkg->yLoc == thePkg->y)
   {
      /* package was dropped at its destination, so
       * remove it from the package list
       */
      removeElement(pkgL, thePkg);
   }
   robot->cpty = robot->cpty + thePkg->weight;
   return 0;
}

void printState(s_world *world, list *robotL, list *homeL, list *pkgL)
{
   printHomeL(homeL);
   printPkgL(pkgL);
   printRobotL(robotL);
   printWorld(world, robotL);
}

void printWorld(s_world *world, list *robotL)
{
   int i;
   printf("\nWORLD ------ Width: %d Height: %d\n", world->width, world->height);
   for(i = world->height + 1 ; i >= 0; i--)
   {
      int j = 0;
      for(j = 0; j < world->width + 2; j++)
      {
         if(robotAt(robotL, j, i))
            printf("X");
         else
            printf("%c", world->map[j][i]);
      }
      printf("\n");
   }
}

void printRobotL(list *robotL)
{
   element *theBot;
   printf("\n--------ROBOT LIST--------------------\n");
   if(headElement(robotL, &theBot) < 0)
   {
      printf("No robots defined\n");
      return;
   }
   while(theBot != NULL)
   {
      element *thePkg;
      printf("ID: %d ", ((s_robot *)theBot->value)->id);
      printf("X: %d ", ((s_robot *)theBot->value)->x);
      printf("Y: %d ", ((s_robot *)theBot->value)->y);
      printf("Money: %d ", ((s_robot *)theBot->value)->money);
      printf("Capacity: %d\n", ((s_robot *)theBot->value)->cpty);
      printf("Packages: ");
      if(headElement(((s_robot *)theBot->value)->pkgs, &thePkg) < 0)
      {
         printf("none\n");
      }
      else
      {
         while(thePkg != NULL)
         {
            printf("%d ", ((s_pkg *)thePkg->value)->id);
            nextElement(((s_robot *)theBot->value)->pkgs, thePkg, &thePkg);
         }
         printf("\n");
      }
      printf("\n");
      nextElement(robotL, theBot, &theBot);
   }
   printf("-------END ROBOT LIST---------------------\n\n");
}

void printHomeL(list *homeL)
{
   element *theHome;
   printf("\n@@@@@@ HOME LIST @@@@@@@@@@@@@@@@@@@@@@@@@@@@\n");
   if(headElement(homeL, &theHome) < 0)
   {
      printf("No Homes defined\n");
   }
   while(theHome != NULL)
   {
      printf("X: %d ", ((s_home *)theHome->value)->x);
      printf("Y: %d", ((s_home *)theHome->value)->y);
      printf("\n");
      nextElement(homeL, theHome, &theHome); 
   }
   printf("@@@@@@@ END HOME LIST @@@@@@@@@@@@@@@@@@@@@@@@@@@\n\n");
}

void printPkgL(list *pkgL)
{
   element *thePkg;
   printf("\n$$$$$$$$$$ PACKAGE LIST $$$$$$$$$$$$$$$$$$$$$$$\n");
   if(headElement(pkgL, &thePkg) < 0)
   {
      printf("No Packages defined\n");
   }
   while(thePkg != NULL)
   {
      printf("ID: %d ", ((s_pkg *)thePkg->value)->id);
      printf("Weight: %d ", ((s_pkg *)thePkg->value)->weight);
      printf("xLoc: %d ", ((s_pkg *)thePkg->value)->xLoc);
      printf("yLoc: %d ", ((s_pkg *)thePkg->value)->yLoc);
      printf("xDest: %d ", ((s_pkg *)thePkg->value)->x);
      printf("yDest: %d ", ((s_pkg *)thePkg->value)->y);
      printf("state: %d ", ((s_pkg *)thePkg->value)->state);
      printf("\n");
      nextElement(pkgL, thePkg, &thePkg);
   }
   printf("$$$$$$$$$$ END PACKAGE LIST $$$$$$$$$$$$$$$$$$$$$\n\n");
}

int botX(list *robotL)
{
   element *head;
   headElement(robotL, &head);
   return ((s_robot *) head->value)->x;
}

int botY(list *robotL)
{
   element *head;
   headElement(robotL, &head);
   return ((s_robot *) head->value)->y;
}

/* returns 1 if yes, 0 if no -1 if error */
int robotAt(list *robotL, int x, int y)
{
   element *head;
   headElement(robotL, &head);
   while(head != NULL)
   {
      if(((s_robot *) head->value)->x == x &&
         ((s_robot *) head->value)->y == y)
      {
         return 1;
      }
      nextElement(robotL, head, &head);
   }
   return 0;
}
